#!/usr/bin/env python
# encoding: utf-8
#
# @Author: José Sánchez-Gallego
# @Date: Nov 3, 2017
# @Filename: spaxel.py
# @License: BSD 3-Clause
# @Copyright: José Sánchez-Gallego


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import inspect
import itertools
import six
import warnings

import numpy as np

import marvin
import marvin.core.core
import marvin.core.exceptions
from marvin.core.exceptions import MarvinError, MarvinUserWarning, MarvinBreadCrumb
import marvin.core.marvin_pickle
import marvin.utils.general.general

import marvin.tools.cube
import marvin.tools.maps
import marvin.tools.modelcube

from marvin.utils.general.structs import FuzzyDict
from marvin.utils.datamodel.dap import datamodel as dap_datamodel
from marvin.utils.datamodel.drp import datamodel as drp_datamodel


breadcrumb = MarvinBreadCrumb()


def spaxel_factory(cls, *args, **kwargs):
    """A factory that returns the right type of spaxel depending on input.

    Based on the input values, determines if the resulting spaxels should be
    binned or unbinned, returning a `.Spaxel` or a `.Bin` respectively.
    This function is intended for overrding the ``__call__`` method in the
    `abc.ABCMeta` metacalass. The reason is that we want `.SpaxelBase` to have
    `abstract methods <abc.abstractmethod>` while also being a factory.
    See `this stack overflow <https://stackoverflow.com/a/5961102>`_ for
    details in the implementation of the ``__call__`` factory pattern.

    It can be used as::

        SpaxelABC = abc.ABCMeta
        SpaxelABC.__call__ = region_factory


        class SpaxelBase(object, metaclass=RegionABC):
            ...

    Note that this will override ``__call__`` everywhere else where
    `abc.ABCMeta` is used, but since it only overrides the default behaviour
    when the input class is `.SpaxelBase`, that should not be a problem.

    """

    Maps = marvin.tools.maps.Maps
    ModelCube = marvin.tools.modelcube.ModelCube

    if cls is not SpaxelBase:
        return type.__call__(cls, *args, **kwargs)

    cube = kwargs.pop('cube', None)
    maps = kwargs.pop('maps', None)
    modelcube = kwargs.pop('modelcube', None)

    plateifu = kwargs.pop('plateifu', (getattr(cube, 'plateifu', None) or
                                       getattr(maps, 'plateifu', None) or
                                       getattr(modelcube, 'plateifu', None)))

    mangaid = kwargs.pop('mangaid', (getattr(cube, 'mangaid', None) or
                                     getattr(maps, 'mangaid', None) or
                                     getattr(modelcube, 'mangaid', None)))

    release = kwargs.pop('release', (getattr(cube, 'release', None) or
                                     getattr(maps, 'release', None) or
                                     getattr(modelcube, 'release', None)))

    spaxel_kwargs = kwargs.copy()
    spaxel_kwargs.update(cube=cube, maps=maps, modelcube=modelcube,
                         plateifu=plateifu, mangaid=mangaid, release=release)

    if not cube and not maps and not modelcube:
        raise MarvinError('no inputs defined.')

    if not maps and not modelcube:
        return Spaxel(*args, **spaxel_kwargs)

    if isinstance(maps, Maps) or isinstance(modelcube, ModelCube):
        bintype = getattr(maps, 'bintype', None) or getattr(modelcube, 'bintype', None)
        spaxel_kwargs.update(bintype=bintype)
        if bintype.binned:
            return Bin(*args, **spaxel_kwargs)
        else:
            return Spaxel(*args, **spaxel_kwargs)

    if maps:
        maps = Maps((maps if maps is not True else None) or plateifu or mangaid,
                    release=release, **kwargs)
        spaxel_kwargs.update(bintype=maps.bintype)
        if maps.bintype.binned:
            return Bin(*args, **spaxel_kwargs)

    if modelcube:
        modelcube = ModelCube((modelcube if modelcube is not True else None) or
                              plateifu or mangaid, release=release, **kwargs)
        spaxel_kwargs.update(bintype=modelcube.bintype)
        if modelcube.bintype.binned:
            return Bin(*args, **spaxel_kwargs)

    return Spaxel(*args, **spaxel_kwargs)

    raise MarvinError('you have reached the end of the SpaxelBase logic. '
                      'This should never happen!')


# Overrides the __call__ method in abc.ABC.
SpaxelABC = abc.ABCMeta
SpaxelABC.__call__ = spaxel_factory


class DataModel(object):
    """A single object that holds the DRP and DAP datamodel."""

    def __init__(self, release):

        self.drp = drp_datamodel[release]
        self.dap = dap_datamodel[release]


class SpaxelBase(six.with_metaclass(SpaxelABC, object)):
    """A base class that contains information about a spaxel.

    This class represents an spaxel with information from the reduced DRP
    spectrum, the DAP maps properties, and the model spectrum from the DAP
    logcube. A `.SpaxelBase` can be initialised with all or only part of that
    information, and either from a file, a database, or remotely via the
    Marvin API.

    The `~marvin.tools.cube.Cube`, `~marvin.tools.maps.Maps` , and
    `~marvin.tools.modelcube.ModelCube` quantities for the spaxel are available
    in ``cube_quantities``, ``maps_quantities``, and ``modelcube_quantities``,
    respectively. For convenience, the quantities can also be accessed directly
    from the `.SpaxelBase` itself (e.g., ``spaxel.emline_gflux_ha_6465``).

    Parameters:
        x,y (int):
            The `x` and `y` coordinates of the spaxel in the cube (0-indexed).
        mangaid (str):
            The mangaid of the cube/maps/modelcube of the spaxel to load.
        plateifu (str):
            The plate-ifu of the cube/maps/modelcube of the spaxel to load
            (either ``mangaid`` or ``plateifu`` can be used, but not both).
        cube (`~marvin.tools.cube.Cube` object or path or bool):
            If ``cube`` is a `~marvin.tools.cube.Cube` object, that
            cube will be used for the `.SpaxelBase` instantiation. This mode
            is mostly intended for `~marvin.utils.general.general.getSpaxel`
            as it significantly improves loading time. Otherwise, ``cube`` can
            be ``True`` (default), in which case a cube will be instantiated
            using the input ``filename``, ``mangaid``, or ``plateifu``. If
            ``cube=False``, no cube will be used and the cube associated
            quantities will not be available. ``cube`` can also be the
            path to the DRP cube to use.
        maps (`~marvin.tools.maps.Maps` object or path or bool)
            As ``cube`` but for the DAP measurements corresponding to the
            spaxel in the `.Maps`.
        modelcube (`marvin.tools.modelcube.ModelCube` object or path or bool)
            As ``maps`` but for the DAP measurements corresponding to the
            spaxel in the `.ModelCube`.
        lazy (bool):
            If ``False``, the spaxel data is loaded on instantiation.
            Otherwise, only the metadata is created. The associated quantities
            can be then loaded by calling `.SpaxelBase.load()`.
        kwargs (dict):
            Arguments to be passed to `.Cube`, `.Maps`, and `.ModelCube`
            when (and if) they are initialised.

    Attributes:
        cube_quantities (`~marvin.utils.general.structs.FuzzyDict`):
            A querable dictionary with the `.Spectrum` quantities
            derived from `.Cube` and matching ``x, y``.
        datamodel (object):
            An object contianing the DRP and DAP datamodels.
        maps_quantities (`~marvin.utils.general.structs.FuzzyDict`):
            A querable dictionary with the `.AnalysisProperty` quantities
            derived from `.Maps` and matching ``x, y``.
        model_quantities (`~marvin.utils.general.structs.FuzzyDict`):
            A querable dictionary with the `.Spectrum` quantities
            derived from `.ModelCube` and matching ``x, y``.
        ra,dec (float):
            Right ascension and declination of the spaxel. Not available until
            the spaxel has been `loaded <.SpaxelBase.load>`.

    """

    def __init__(self, x, y, mangaid=None, plateifu=None,
                 cube=True, maps=True, modelcube=True, lazy=False, **kwargs):

        self.cube_quantities = FuzzyDict({})
        self.maps_quantities = FuzzyDict({})
        self.modelcube_quantities = FuzzyDict({})

        self._cube = cube
        self._maps = maps
        self._modelcube = modelcube

        self._plateifu = plateifu
        self._mangaid = mangaid

        self.kwargs = kwargs

        if not self._cube and not self._maps and not self._modelcube:
            raise MarvinError('either cube, maps, or modelcube must be True or '
                              'a Marvin Cube, Maps, or ModelCube object must be specified.')

        self._parent_shape = None

        # drop breadcrumb
        breadcrumb.drop(message='Initializing MarvinSpaxel {0}'.format(self.__class__),
                        category=self.__class__)

        self.x = int(x)
        self.y = int(y)

        self.loaded = False
        self.datamodel = None

        if lazy is False:
            self.load()

    def __dir__(self):

        class_members = list(list(zip(*inspect.getmembers(self.__class__)))[0])
        instance_attr = list(self.__dict__.keys())

        items = self.cube_quantities.__dir__()
        items += self.maps_quantities.__dir__()
        items += self.modelcube_quantities.__dir__()
        items += class_members + instance_attr

        return sorted(items)

    def __getattr__(self, value):

        _getattr = super(SpaxelBase, self).__getattribute__

        if value in _getattr('cube_quantities'):
            return _getattr('cube_quantities')[value]
        if value in _getattr('maps_quantities'):
            return _getattr('maps_quantities')[value]
        if value in _getattr('modelcube_quantities'):
            return _getattr('modelcube_quantities')[value]

        return super(SpaxelBase, self).__getattribute__(value)

    @abc.abstractmethod
    def __repr__(self):

        return '<SpaxelBase>'

    def _check_versions(self, attr):
        """Checks that all input object have the same versions.

        Runs sanity checks to make sure that ``attr`` has the same value
        in the input `.Cube`, `.Maps`, and `.ModelCube`. Returns the value
        for the attribute or ``None`` if the attribute does not exist.

        """

        out_value = None

        inputs = []
        for obj in [self._cube, self._maps, self._modelcube]:
            if obj is not None and not isinstance(obj, bool):
                inputs.append(obj)

        if len(inputs) == 1:
            return getattr(inputs[0], attr, None)

        for obj_a, obj_b in itertools.combinations(inputs, 2):
            if hasattr(obj_a, attr) and hasattr(obj_b, attr):
                assert getattr(obj_a, attr) == getattr(obj_b, attr), \
                    'inconsistent {!r} between {!r} and {!r}'.format(attr, obj_a, obj_b)
            out_value = getattr(obj_a, attr, None) or getattr(obj_b, attr, None)

        return out_value

    def _set_radec(self):
        """Calculates ra and dec for this spaxel."""

        self.ra = None
        self.dec = None

        for obj in [self._cube, self._maps, self._modelcube]:
            if hasattr(obj, 'wcs'):
                if obj.wcs.naxis == 2:
                    self.ra, self.dec = obj.wcs.wcs_pix2world([[self.x, self.y]], 0)[0]
                elif obj.wcs.naxis == 3:
                    self.ra, self.dec, __ = obj.wcs.wcs_pix2world([[self.x, self.y, 0]], 0)[0]

    def load(self, force=None):
        """Loads the spaxel data.

        Loads the spaxel data for cubes/maps/modelcubes.  By default attempts
        to load whatever is specified when spaxels are instantianted from other
        Marvin Tools.  Can manually force load a data type with the force
        keyword.

        Parameters:
            force ({cube|maps|models}):
                Str datatype to force load
        """

        if self.loaded and force is None:
            warnings.warn('already loaded', MarvinUserWarning)
            return

        # setup what will be loaded
        self._setup_load(force=force)

        self._load_cube()
        self._load_maps()
        self._load_modelcube()

        self._set_radec()
        self.loaded = True

        for attr in ['mangaid', 'plateifu', 'release', 'bintype', 'template']:
            self._check_versions(attr)

        self._plateifu = self.plateifu
        self._mangaid = self.mangaid

        self.datamodel = DataModel(self.release)

    def save(self, path, overwrite=False):
        """Pickles the spaxel to a file.

        Parameters:
            path (str):
                The path of the file to which the `.Spaxel` will be saved.
                Unlike for other Marvin Tools that derive from
                `~marvin.core.core.MarvinToolsClass`, ``path`` is
                mandatory for `.Spaxel.save` as there is no default path for a
                given spaxel.
            overwrite (bool):
                If True, and the ``path`` already exists, overwrites it.
                Otherwise it will fail.

        Returns:
            path (str):
                The realpath to which the file has been saved.

        """

        return marvin.core.marvin_pickle.save(self, path=path, overwrite=overwrite)

    @classmethod
    def restore(cls, path, delete=False):
        """Restores a Spaxel object from a pickled file.

        If ``delete=True``, the pickled file will be removed after it has been
        unplickled. Note that, for objects with ``data_origin='file'``, the
        original file must exists and be in the same path as when the object
        was first created.

        """

        return marvin.core.marvin_pickle.restore(path, delete=delete)

    def _load_cube(self):
        """Loads the cube and the associated quantities."""

        if self._cube is False or self._cube is None:
            self._cube = None
            return
        elif not isinstance(self._cube, marvin.tools.cube.Cube):
            self._cube = self.getCube()

        self._parent_shape = self._cube._shape

        self.cube_quantities = self._cube._get_spaxel_quantities(self.x, self.y)

    def _load_maps(self):
        """Loads the cube and the properties."""

        if self._maps is False or self._maps is None:
            self._maps = None
            return
        elif not isinstance(self._maps, marvin.tools.maps.Maps):
            self._maps = self.getMaps()

        self._parent_shape = self._maps._shape

        self.maps_quantities = self._maps._get_spaxel_quantities(self.x, self.y)

    def _load_modelcube(self):
        """Loads the modelcube and associated arrays."""

        if self._modelcube is False or self._modelcube is None:
            self._modelcube = None
            return
        elif not isinstance(self._modelcube, marvin.tools.modelcube.ModelCube):

            if self.release == 'MPL-4':
                warnings.warn('ModelCube cannot be instantiated for MPL-4.', MarvinUserWarning)
                self._modelcube = None
                return

            self._modelcube = self.getModelCube()

        self._parent_shape = self._modelcube._shape

        self.modelcube_quantities = self._modelcube._get_spaxel_quantities(self.x, self.y)

    def _setup_load(self, force=None):
        ''' Setup the loading for cube/maps/models

        Can manually force load a data type.

        Parameters:
            force ({cube|maps|models}):
                Str datatype to force load
        '''

        # get what is currently loaded
        flag = self._get_loaded()

        # only do something if force to set.  Otherwise do the original nothing.
        if force:
            assert force in ['cube', 'maps', 'models'], 'force can only be cube, maps, or models'

            if force in flag:
                warnings.warn('{0} is already loaded'.format(force), MarvinUserWarning)
            else:
                if force == 'cube' and not self._cube:
                    self._cube = True
                elif force == 'maps' and not self._maps:
                    self._maps = True
                elif force == 'models' and not self._modelcube:
                    self._modelcube = True
                else:
                    # do nothing
                    pass

    def _get_loaded(self):
        ''' Get a flag to indicate what is or is not loaded '''

        dataloads = [self._cube, self._maps, self._modelcube]
        flags = np.array(['cube', 'maps', 'models'])
        load_idx = np.where(dataloads)[0]
        flag = '/'.join(flags[load_idx])

        return flag

    def getCube(self):
        """Returns the associated `~marvin.tools.cube.Cube`"""

        if isinstance(self._cube, marvin.tools.cube.Cube):
            return self._cube

        cube_kwargs = self.kwargs.copy()
        cube_kwargs.pop('bintype', None)
        cube_kwargs.pop('template', None)

        return marvin.tools.cube.Cube(
            (self._cube if self._cube is not True else None) or self._plateifu or self._mangaid,
            **cube_kwargs)

    def getMaps(self):
        """Returns the associated `~marvin.tools.maps.Maps`"""

        if isinstance(self._maps, marvin.tools.maps.Maps):
            return self._maps

        maps_kwargs = self.kwargs.copy()

        bintype = maps_kwargs.pop('bintype', None) or self.bintype
        template = maps_kwargs.pop('template', None) or self.template
        release = maps_kwargs.pop('release', None) or self.release

        return marvin.tools.maps.Maps(
            (self._maps if self._maps is not True else None) or self._plateifu or self._mangaid,
            bintype=bintype, template=template, release=release, **maps_kwargs)

    def getModelCube(self):
        """Returns the associated `~marvin.tools.modelcube.ModelCube`"""

        if isinstance(self._modelcube, marvin.tools.modelcube.ModelCube):
            return self._modelcube

        modelcube_kwargs = self.kwargs.copy()

        bintype = modelcube_kwargs.pop('bintype', None) or self.bintype
        template = modelcube_kwargs.pop('template', None) or self.template
        release = modelcube_kwargs.pop('release', None) or self.release

        return marvin.tools.modelcube.ModelCube(
            ((self._modelcube if self._modelcube is not True else None) or
             self._plateifu or self._mangaid),
            bintype=bintype, template=template, release=release, **modelcube_kwargs)

    @property
    def plateifu(self):
        """Returns the plateifu."""

        return self._check_versions('plateifu')

    @property
    def mangaid(self):
        """Returns the mangaid."""

        return self._check_versions('mangaid')

    @property
    def release(self):
        """Returns the release."""

        return self._check_versions('release')

    @property
    def bintype(self):
        """Returns the bintype."""

        return self._check_versions('bintype')

    @property
    def template(self):
        """Returns the template."""

        return self._check_versions('template')

    @property
    def manga_target1(self):
        """Return MANGA_TARGET1 flag."""
        return self.datamodel.drp.bitmasks['MANGA_TARGET1']

    @property
    def manga_target2(self):
        """Return MANGA_TARGET2 flag."""
        return self.datamodel.drp.bitmasks['MANGA_TARGET2']

    @property
    def manga_target3(self):
        """Return MANGA_TARGET3 flag."""
        return self.datamodel.drp.bitmasks['MANGA_TARGET3']

    @property
    def target_flags(self):
        """Bundle MaNGA targeting flags."""
        return [self.manga_target1, self.manga_target2, self.manga_target3]

    @property
    def quality_flags(self):
        """Bundle Cube DRP3QUAL and Maps DAPQUAL flags."""

        drp3qual = self.datamodel.drp.bitmasks['MANGA_DRP3QUAL']
        cube = self.getCube()
        drp3qual.mask = int(cube.header['DRP3QUAL'])
        qual_flags = [drp3qual]

        if self.release != 'MPL-4':
            qual_flags.append(self.datamodel.dap.bitmasks['MANGA_DAPQUAL'])

        return qual_flags


class Spaxel(SpaxelBase):
    """A class representing an unbinned spaxel.

    This subclass of `.SpaxelBase` represents an spaxel belonging to an
    unbinned `~marvin.tools.maps.Maps` and `~marvin.tools.modelcube.ModelCube`.
    If initialised directly, a `.Spaxel` will fail if the input data correspond
    to a binned maps or modelcube object.

    Refer to the documentation of `.SpaxelBase` for information about the valid
    parameters and methods.

    """

    def __init__(self, *args, **kwargs):

        super(Spaxel, self).__init__(*args, **kwargs)

        if isinstance(self._maps, marvin.tools.maps.Maps):
            assert self._maps.is_binned() is False, 'a Spaxel cannot be binned.'

        if isinstance(self._modelcube, marvin.tools.modelcube.ModelCube):
            assert self._modelcube.is_binned() is False, 'a Spaxel cannot be binned.'

    def __repr__(self):
        """Spaxel representation."""

        if not self.loaded:
            return '<Marvin Spaxel (x={0.x:d}, y={0.y:d}, loaded=False)'.format(self)

        # Gets the coordinates relative to the centre of the cube/maps.
        y_mid, x_mid = np.array(self._parent_shape) / 2.
        x_centre = int(self.x - x_mid)
        y_centre = int(self.y - y_mid)

        return ('<Marvin Spaxel (plateifu={0.plateifu}, x={0.x:d}, y={0.y:d}; '
                'x_cen={1:d}, y_cen={2:d}, loaded={3})>'.format(self, x_centre, y_centre,
                                                                self._get_loaded()))


class Bin(SpaxelBase):
    """A class that represents a bin.

    Attributes
    ----------
    spaxels : list
        A list of `.Spaxel` instances that provides access to the unbinned
        spaxels that are associated with the bin.

    """

    def __init__(self, *args, **kwargs):

        self.binid = None
        self.spaxels = None

        super(Bin, self).__init__(*args, **kwargs)

        if isinstance(self._maps, marvin.tools.maps.Maps):
            assert self._maps.is_binned() is True, 'a Spaxel cannot be unbinned.'

        if isinstance(self._modelcube, marvin.tools.modelcube.ModelCube):
            assert self._modelcube.is_binned() is True, 'a Spaxel cannot be unbinned.'

        assert (isinstance(self._maps, marvin.tools.maps.Maps) or
                isinstance(self._modelcube, marvin.tools.modelcube.ModelCube)), \
            'a Bin must have a Maps or a ModelCube.'

    def load(self):
        """Loads quantities and spaxels."""

        super(Bin, self).load()
        self._create_spaxels()

    def __repr__(self):
        """Spaxel representation."""

        if not self.loaded:
            return '<Marvin Bin (x={0.x:d}, y={0.y:d}, loaded=False)'.format(self)

        # Gets the coordinates relative to the centre of the cube/maps.
        y_mid, x_mid = np.array(self._parent_shape) / 2.
        x_centre = int(self.x - x_mid)
        y_centre = int(self.y - y_mid)

        return ('<Marvin Bin (plateifu={0.plateifu}, x={0.x:d}, y={0.y:d}; '
                'x_cen={1:d}, y_cen={2:d}, n_spaxels={3}, loaded={4})>'.format(self, x_centre,
                                                                               y_centre, len(self.spaxels),
                                                                               self._get_loaded()))

    def _create_spaxels(self):
        """Returns a list of the unbinned spaxels associated with this bin."""

        if self._maps:
            binid_map = self._maps.get_binid()
        elif self._modelcube:
            binid_map = self._modelcube.get_binid()

        # TODO: this may be an overkill and an extra API call. Remove it?
        if self._maps and self._modelcube:
            assert np.all(self._maps.get_binid() == self._modelcube.get_binid()), \
                'inconsistent binid arrays between Maps and ModelCube.'

        self.binid = binid_map[self.y, self.x]

        if self.binid < 0:
            raise MarvinError('coordinates ({}, {}) do not correspond to a valid binid.'
                              .format(self.x, self.y))

        spaxel_coords = zip(*np.where(binid_map == self.binid))
        self.spaxels = []

        for ii, jj in spaxel_coords:
            self.spaxels.append(Spaxel(x=jj, y=ii, plateifu=self.plateifu, release=self.release,
                                       cube=self._cube, maps=True if self._maps else False,
                                       modelcube=True if self._modelcube else False,
                                       bintype=None, template=self.template, lazy=True))

    def load_all(self):
        """Loads all the spaxels."""

        for spaxel in self.spaxels:
            spaxel.load()



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-12 18:18:54
# @Last modified by:   Brian Cherinka
# @Last modified time: 2017-07-31 12:07:88

from __future__ import print_function, division, absolute_import
from marvin.tools.query import Query
from marvin.tools.results import Results, ResultSet, marvintuple
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.spaxel import Spaxel
from marvin.tools.modelcube import ModelCube
from marvin.utils.datamodel.query.base import ParameterGroup, query_params
from marvin import config
from marvin.core.exceptions import MarvinError
from collections import OrderedDict, namedtuple
import pytest
import copy
import six
import pandas as pd
import json


myplateifu = '8485-1901'
cols = ['cube.mangaid', 'cube.plate', 'cube.plateifu', 'ifu.name', 'nsa.z']
remotecols = [u'mangaid', u'plate', u'plateifu', u'ifu_name', u'z']


@pytest.fixture()
def limits(results):
    data = results.expdata['queries'][results.searchfilter]
    count = 10 if data['count'] >= 10 else data['count']
    return (10, count)


@pytest.fixture()
def results(query, request):
    searchfilter = request.param if hasattr(request, 'param') else 'nsa.z < 0.1 and cube.plate==8485'
    q = Query(searchfilter=searchfilter, mode=query.mode, limit=10, release=query._release)
    r = q.run()
    r.expdata = query.expdata
    yield r
    r = None


@pytest.fixture()
def columns(results):
    ''' returns the local or remote column syntax '''
    return ParameterGroup('Columns', cols)


class TestResultsColumns(object):

    def test_check_cols(self, results, columns):
        assert results.columns.full == columns.full
        assert results.columns.remote == columns.remote

    @pytest.mark.parametrize('col, errmsg', [('plate', 'plate is too ambiguous.  Did you mean one of')])
    def test_fail(self, results, col, errmsg):
        with pytest.raises(KeyError) as cm:
            col in results.columns
        assert cm.type == KeyError
        assert errmsg in str(cm.value)

    @pytest.mark.parametrize('results', [('nsa.z < 0.1 and haflux > 25')], indirect=True)
    @pytest.mark.parametrize('colmns, pars', [(['spaxelprop.x', 'spaxelprop.y', 'emline_gflux_ha_6564', 'bintype.name', 'template.name'],
                                               ['x', 'y', 'emline_gflux_ha_6564', 'bintype_name', 'template_name'])])
    def test_check_withadded(self, results, colmns, pars, columns):
        newcols = cols + colmns
        assert set(results._params) == set(newcols)

        newcols = copy.copy(columns)
        newcols.extend(ParameterGroup('Columns', colmns))
        assert set(results.columns.full) == set(newcols.full)
        assert set(results.columns.remote) == set(newcols.remote)

    @pytest.mark.parametrize('results', [('nsa.z < 0.1 and haflux > 25')], indirect=True)
    @pytest.mark.parametrize('full, name', [('nsa.z', 'z'), ('cube.plateifu', 'plateifu'),
                                            ('cube.mangaid', 'mangaid'),
                                            ('spaxelprop.emline_gflux_ha_6564', 'haflux')])
    def test_colnames(self, results, full, name):
        cols = results.columns
        assert full in cols
        assert name in cols
        col = cols[name]
        assert col.full == full


class TestMarvinTuple(object):

    def test_create(self, results):
        data = results.results[0].__dict__
        mt = marvintuple('ResultRow', data.keys())
        assert mt is not None
        assert hasattr(mt, 'mangaid')
        assert hasattr(mt, 'plateifu')

        row = mt(**data)
        assert row.plate == data['plate']
        assert row.mangaid == data['mangaid']
        assert row.plateifu == data['plateifu']

    def test_add(self, results):
        data = results.results[0].__dict__
        mt = marvintuple('ResultRow', 'mangaid, plate, plateifu, ifu_name')
        mt1 = marvintuple('ResultRow', 'z')

        row = mt(**{k: v for k, v in data.items() if k in ['mangaid', 'plate', 'plateifu', 'ifu_name']})
        row1 = mt1(**{k: v for k, v in data.items()})
        new_row = row + row1
        assert new_row is not None
        cols = ['mangaid', 'plate', 'plateifu', 'z']
        assert set(cols).issubset(set(new_row.__dict__.keys()))
        assert row.__dict__ <= new_row.__dict__


class TestResultSet(object):

    def test_list(self, results):
        reslist = results.results.to_list()
        assert isinstance(reslist, list)

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_sort(self, results):
        redshift = results.expdata['queries']['nsa.z < 0.1']['sorted']['1'][-1]
        results.getAll()
        results.results.sort('z')
        assert results.results['z'][0] == redshift

    def test_add(self, results):
        res = results.results
        res1 = res
        newres = res + res1
        newres1 = res1 + res
        assert res.columns.full == newres.columns.full
        assert newres == newres1


class TestResultsMisc(object):

    def test_showQuery(self, results):
        x = results.showQuery()
        assert isinstance(x, six.string_types)


class TestResultsOutput(object):

    def test_tofits(self, results, temp_scratch):
        file = temp_scratch.join('test_results.fits')
        results.toFits(filename=str(file), overwrite=True)
        assert file.check(file=1, exists=1) is True

    def test_tocsv(self, results, temp_scratch):
        file = temp_scratch.join('test_results.csv')
        results.toCSV(filename=str(file), overwrite=True)
        assert file.check(file=1, exists=1) is True

    def test_topandas(self, results):
        df = results.toDF()
        assert isinstance(df, pd.core.frame.DataFrame)

    def test_tojson(self, results):
        res = results.toJson()
        assert isinstance(res, six.string_types)
        json_res = json.loads(res)
        assert isinstance(json_res, list)


class TestResultsGetParams(object):

    def test_get_attribute(self, results, columns):
        res = results.results[0]
        assert isinstance(results.results, ResultSet) is True
        for i, name in enumerate(columns):
            assert res[i] == res.__getattribute__(name.remote)

    @pytest.mark.parametrize('col',
                             [(c) for c in cols],
                             ids=['cube.mangaid', 'cube.plate', 'cube.plateifu', 'ifu.name', 'nsa.z'])
    def test_get_list(self, results, col):
        assert col in results.columns
        obj = results.getListOf(col)
        assert obj is not None
        assert isinstance(obj, list) is True
        json_obj = results.getListOf(col, to_json=True)
        assert isinstance(json_obj, six.string_types)

    @pytest.mark.parametrize('results',
                             [('nsa.z < 0.1 and haflux > 25'),
                              ('nsa.z < 0.1'),
                              ('haflux > 25')], indirect=True)
    def test_get_list_all(self, results):
        q = Query(searchfilter=results.searchfilter, mode=results.mode, limit=1,
                  release=results._release, returnparams=results.returnparams)
        r = q.run(start=0, end=1)
        assert r.count == 1
        mangaids = r.getListOf('mangaid', return_all=True)
        assert len(mangaids) == r.totalcount
        assert len(mangaids) == results.expdata['queries'][results.searchfilter]['count']

    @pytest.mark.parametrize('ftype', [('dictlist'), ('listdict')])
    @pytest.mark.parametrize('name', [(None), ('mangaid'), ('z')], ids=['noname', 'mangaid', 'z'])
    def test_get_dict(self, results, ftype, name):
        output = results.getDictOf(name, format_type=ftype)

        if ftype == 'listdict':
            assert isinstance(output, list) is True
            assert isinstance(output[0], dict) is True
            if name is not None:
                assert set([name]) == set(list(output[0]))
            else:
                assert set(remotecols) == set(output[0])
        elif ftype == 'dictlist':
            assert isinstance(output, dict) is True
            assert isinstance(list(output.values())[0], list) is True
            if name is not None:
                assert set([name]) == set(list(output.keys()))
            else:
                assert set(remotecols) == set(output)

        json_obj = results.getDictOf(name, format_type=ftype, to_json=True)
        assert isinstance(json_obj, six.string_types)

    def test_get_dict_all(self, results):
        output = results.getDictOf('mangaid', return_all=True)
        assert len(output) == results.totalcount


class TestResultsSort(object):

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_sort(self, results, limits):
        results.sort('z')
        limit, count = limits
        data = results.expdata['queries'][results.searchfilter]['sorted']
        assert tuple(data['1']) == results.results[0]
        assert tuple(data[str(count)]) == results.results[count - 1]


class TestResultsPaging(object):

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_check_counts(self, results, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        results.sort('z')
        limit, count = limits
        data = results.expdata['queries'][results.searchfilter]
        assert results.totalcount == data['count']
        assert results.count == count
        assert len(results.results) == count
        assert results.limit == limit
        assert results.chunk == limit

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('chunk, rows',
                             [(10, None),
                              (20, (10, 21))],
                             ids=['defaultchunk', 'chunk20'])
    def test_get_next(self, results, chunk, rows, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        limit, count = limits
        results.sort('z')
        results.getNext(chunk=chunk)
        data = results.expdata['queries'][results.searchfilter]['sorted']
        if results.count == results.totalcount:
            assert results.results[0] == tuple(data['1'])
            assert len(results.results) == count
        else:
            assert results.results[0] == tuple(data['11'])
            assert len(results.results) == chunk
            if rows:
                assert results.results[rows[0]] == tuple(data[str(rows[1])])

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('index, chunk, rows',
                             [(30, 10, [(0, 21)]),
                              (45, 20, [(5, 31), (10, 36), (15, 41)])],
                             ids=['defaultchunk', 'chunk20'])
    def test_get_prev(self, results, index, chunk, rows, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        limit, count = limits
        results.sort('z')
        results.getSubset(index, limit=chunk)
        results.getPrevious(chunk=chunk)
        data = results.expdata['queries'][results.searchfilter]['sorted']
        if results.count == results.totalcount:
            assert results.results[0] == tuple(data['1'])
            assert len(results.results) == count
        elif results.count == 0:
            assert len(results.results) == 0
        else:
            assert len(results.results) == chunk
            if rows:
                for row in rows:
                    assert results.results[row[0]] == tuple(data[str(row[1])])

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    @pytest.mark.parametrize('index, chunk, rows',
                             [(35, 10, [(0, 36)]),
                              (30, 20, [(0, 31), (15, 46)])],
                             ids=['defaultchunk', 'chunk20'])
    def test_get_set(self, results, index, chunk, rows, limits):
        if results.mode == 'local':
            pytest.skip('skipping now due to weird issue with local results not same as remote results')
        limit, count = limits
        results.sort('z')
        results.getSubset(index, limit=chunk)
        data = results.expdata['queries'][results.searchfilter]['sorted']
        if results.count == results.totalcount:
            assert results.results[0] == tuple(data['1'])
            assert len(results.results) == count
        elif results.count == 0:
            assert len(results.results) == 0
        else:
            assert len(results.results) == chunk
            if rows:
                for row in rows:
                    assert results.results[row[0]] == tuple(data[str(row[1])])

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_extend_set(self, results):
        res = results.getSubset(0, limit=1)
        assert results.count == 1
        assert len(results.results) == 1
        results.extendSet(start=1, chunk=2)
        setcount = 3 if results.count > 3 else results.count
        assert results.count == setcount
        assert len(results.results) == setcount

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_loop(self, results):
        res = results.getSubset(0, limit=1)
        assert results.count == 1
        results.loop(chunk=500)
        assert results.count == results.expdata['queries'][results.searchfilter]['count']
        assert results.count == results.totalcount

    @pytest.mark.parametrize('results', [('nsa.z < 0.1')], indirect=True)
    def test_get_all(self, results):
        res = results.getAll()
        assert results.count == results.totalcount


class TestResultsPickling(object):

    def test_pickle_save(self, results, temp_scratch):
        file = temp_scratch.join('test_results.mpf')
        path = results.save(str(file), overwrite=True)
        assert file.check() is True

    def test_pickle_restore(self, results, temp_scratch):
        file = temp_scratch.join('test_results.mpf')
        path = results.save(str(file), overwrite=True)
        assert file.check() is True
        r = Results.restore(str(file))
        assert r.searchfilter == results.searchfilter


class TestResultsConvertTool(object):

    @pytest.mark.parametrize('results', [('nsa.z < 0.1 and haflux > 25')], indirect=True)
    @pytest.mark.parametrize('objtype, tool',
                             [('cube', Cube), ('maps', Maps), ('spaxel', Spaxel),
                              ('modelcube', ModelCube)])
    def test_convert_success(self, results, objtype, tool, exporigin):
        if config.release == 'MPL-4' and objtype == 'modelcube':
            pytest.skip('no modelcubes in mpl-4')

        results.convertToTool(objtype, limit=1)
        assert results.objects is not None
        assert isinstance(results.objects, list) is True
        assert isinstance(results.objects[0], tool) is True
        if objtype != 'spaxel':
            assert results.mode == results.objects[0].mode

    @pytest.mark.parametrize('objtype, error, errmsg',
                             [('modelcube', MarvinError, "ModelCube requires at least dapver='2.0.2'"),
                              ('spaxel', AssertionError, 'Parameters must include spaxelprop.x and y in order to convert to Marvin Spaxel')],
                             ids=['mcminrelease', 'nospaxinfo'])
    def test_convert_failures(self, results, objtype, error, errmsg):
        if config.release > 'MPL-4' and objtype == 'modelcube':
            pytest.skip('modelcubes in post mpl-4')

        with pytest.raises(error) as cm:
            results.convertToTool(objtype, limit=1)
        assert cm.type == error
        assert errmsg in str(cm.value)





# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-12 18:41:25
# @Last modified by:   andrews
# @Last modified time: 2018-03-02 17:03:81

from __future__ import print_function, division, absolute_import
import pytest


@pytest.fixture()
def nsa_target(maindb, galaxy):
    nsa = maindb.session.query(maindb.sampledb.NSA).join(maindb.sampledb.MangaTargetToNSA,
                                                         maindb.sampledb.MangaTarget).\
        filter(maindb.sampledb.MangaTarget.mangaid == galaxy.mangaid).one()
    yield (nsa, galaxy.plateifu)
    nsa = None


class TestSampleDB(object):

    @pytest.mark.parametrize('plateifu, expected',
                             [('8485-1901', {'u': 18.69765903,
                                             'g': 17.45450578,
                                             'r': 16.80842176,
                                             'i': 16.43652498,
                                             'z': 16.20534984}),
                              ('7443-12701', {'u': 17.07501837,
                                              'g': 15.57770095,
                                              'r': 14.95969099,
                                              'i': 14.63861064,
                                              'z': 14.44369601})])
    def test_elpetro_mag(self, nsa_target, plateifu, expected):
        nsa_target, galaxy_plateifu = nsa_target

        if galaxy_plateifu != plateifu:
            pytest.skip('Skip non-matching plateifus.')

        for band, value in expected.items():
            assert getattr(nsa_target, 'elpetro_mag_{0}'.format(band)) == pytest.approx(value)

    @pytest.mark.parametrize('plateifu, expected',
                             [('8485-1901', {('u', 'g'): 1.24315324,
                                             ('g', 'r'): 0.64608403,
                                             ('r', 'i'): 0.37189678,
                                             ('i', 'z'): 0.23117514}),
                              ('7443-12701', {('u', 'g'): 1.49731742,
                                              ('g', 'r'): 0.61800996,
                                              ('r', 'i'): 0.32108036,
                                              ('i', 'z'): 0.19491463})])
    def test_elpetro_colour(self, nsa_target, plateifu, expected):
        nsa_target, galaxy_plateifu = nsa_target

        if galaxy_plateifu != plateifu:
            pytest.skip('Skip non-matching plateifus.')

        for bands, value in expected.items():
            bandA, bandB = bands

            assert nsa_target.elpetro_colour(bandA, bandB) == pytest.approx(value)

            elpetro_mag_colour = getattr(nsa_target, 'elpetro_mag_{0}_{1}'.format(bandA, bandB))
            assert elpetro_mag_colour == pytest.approx(value)

    @pytest.mark.parametrize('plateifu, expected',
                             [('8485-1901', {('u', 'g'): 1.1655902862549006,
                                             ('g', 'r'): 0.5961246490479013,
                                             ('r', 'i'): 0.3375816345214986,
                                             ('i', 'z'): 0.20068740844720168}),
                              ('7443-12701', {('u', 'g'): 1.3728961944580007,
                                              ('g', 'r'): 0.5836753845213991,
                                              ('r', 'i'): 0.27035522460939987,
                                              ('i', 'z'): 0.1656112670899006})])
    def test_elpetro_absmag_colour(self, nsa_target, plateifu, expected):
        nsa_target, galaxy_plateifu = nsa_target

        if galaxy_plateifu != plateifu:
            pytest.skip('Skip non-matching plateifus.')

        for bands, value in expected.items():
            bandA, bandB = bands

            assert nsa_target.elpetro_absmag_colour(bandA, bandB) == pytest.approx(value)

            colour = 'elpetro_absmag_{0}_{1}'.format(bandA, bandB)
            elpetro_absmag_colour = getattr(nsa_target, colour)
            assert elpetro_absmag_colour == pytest.approx(value)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-08 16:31:45
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-09 10:20:35

from __future__ import print_function, division, absolute_import
import pytest
from sdss_access.path import Path


class TestPath(object):

    def test_missing_keys(self, path):
        with pytest.raises(KeyError) as cm:
            full = path.full('mangacube')
        assert 'Missing required keyword arguments:' in str(cm.value)

    def test_public(self):
        path = Path()
        url = path.url('mangacube', drpver='v2_3_1', plate=8485, ifu='1901')
        assert 'mangawork' in url

        release = 'dr14'
        path = Path(public=True, release=release)
        url = path.url('mangacube', drpver='v2_3_1', plate=8485, ifu='1901')
        assert release in url




#!/usr/bin/env python
# encoding: utf-8
#
# setup.py
#


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from setuptools import setup, find_packages

import os
import argparse
import sys
import distutils.dir_util

NAME = 'sdss-tree'
VERSION = '2.15.5dev'
RELEASE = 'dev' in VERSION


def run(packages, install_requires):

    setup(name=NAME,
          version=VERSION,
          license='BSD3',
          description='Description of your project.',
          long_description=open('README.rst').read(),
          author='Brian Cherinka',
          author_email='bcherin1@jhu.edu',
          keywords='astronomy software',
          url='https://github.com/sdss/tree',
          include_package_data=True,
          packages=packages,
          install_requires=install_requires,
          package_dir={'': 'python'},
          scripts=[],
          classifiers=[
              'Development Status :: 4 - Beta',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: BSD License',
              'Natural Language :: English',
              'Operating System :: OS Independent',
              'Programming Language :: Python',
              'Programming Language :: Python :: 2.6',
              'Programming Language :: Python :: 2.7',
              'Topic :: Documentation :: Sphinx',
              'Topic :: Software Development :: Libraries :: Python Modules',
          ],
          )


def get_requirements(opts):
    ''' Get the proper requirements file based on the optional argument '''

    if opts.dev:
        name = 'requirements_dev.txt'
    elif opts.doc:
        name = 'requirements_doc.txt'
    else:
        name = 'requirements.txt'

    requirements_file = os.path.join(os.path.dirname(__file__), name)
    install_requires = [line.strip().replace('==', '>=') for line in open(requirements_file)
                        if not line.strip().startswith('#') and line.strip() != '']
    return install_requires


def remove_args(parser):
    ''' Remove custom arguments from the parser '''

    arguments = []
    for action in list(parser._get_optional_actions()):
        if '--help' not in action.option_strings:
            arguments += action.option_strings

    for arg in arguments:
        if arg in sys.argv:
            sys.argv.remove(arg)

if __name__ == '__main__':

    # Custom parser to decide whether which requirements to install
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]))
    parser.add_argument('-d', '--dev', dest='dev', default=False, action='store_true', help='Install all packages for development')
    parser.add_argument('-o', '--doc', dest='doc', default=False, action='store_true', help='Install only core + documentation packages')

    # We use parse_known_args because we want to leave the remaining args for distutils
    args = parser.parse_known_args()[0]

    # Copy data files into the python/tree/data directory
    tmp = distutils.dir_util.copy_tree('data', 'python/tree/data')

    # Get the proper requirements file
    install_requires = get_requirements(args)

    # Now we remove all our custom arguments to make sure they don't interfere with distutils
    remove_args(parser)

    # Have distutils find the packages
    packages = find_packages(where='python')

    # Runs distutils
    run(packages, install_requires)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-02-19 23:37:44
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-02-20 00:24:45

from __future__ import print_function, division, absolute_import
import os
import pytest
from brain.utils.general import getDbMachine, compress_data, uncompress_data


class TestGetDbMachine(object):
    ''' '''

    @pytest.mark.parametrize('expname, hostname',
                             [('local', 'manga'),
                              ('local', ''),
                              ('utah', 'sas-vm.sdss.org'),
                              ('utah', ''),
                              ('jhu', 'jhu.edu'),
                              (None, 'badmachine.org')],
                             ids=['local', 'local-2', 'sasvm', 'utah', 'jhu', 'none'])
    def test_servers(self, monkeypatch, expname, hostname):
        ''' test correct db name is returned '''

        monkeypatch.delitem(os.environ, 'MANGA_LOCALHOST')
        monkeypatch.setitem(os.environ, 'HOSTNAME', hostname)
        if expname == 'utah' and not hostname:
            monkeypatch.setitem(os.environ, 'UUFSCELL', 'kingspeak.peaks')
        if expname == 'local' and not hostname:
            monkeypatch.setitem(os.environ, 'MANGA_LOCALHOST', '1')

        db = getDbMachine()
        assert db == expname


class TestCompression(object):
    ''' '''

    data = {'a': 1, 'b': 2, 'c': 3}

    @pytest.mark.parametrize('comptype, expcomp',
                             [('json', '{"a": 1, "c": 3, "b": 2}'),
                              ('msgpack', '\x83\xc4\x01a\x01\xc4\x01c\x03\xc4\x01b\x02')],
                             ids=['json', 'msgpack'])
    def test_compression(self, comptype, expcomp):
        comp = compress_data(self.data, compress_with=comptype)
        assert comp == expcomp

    @pytest.mark.parametrize('comptype, expcomp',
                             [('json', '{"a": 1, "c": 3, "b": 2}'),
                              ('msgpack', '\x83\xc4\x01a\x01\xc4\x01c\x03\xc4\x01b\x02')],
                             ids=['json', 'msgpack'])
    def test_uncompression(self, comptype, expcomp):
        comp = compress_data(self.data, compress_with=comptype)
        uncomp = uncompress_data(comp, uncompress_with=comptype)
        assert uncomp == self.data



